
from typing import List, Optional, NamedTuple, Literal
from dataclasses import dataclass
from math import floor

from Bio.Seq import Seq
from Bio.SeqRecord import SeqRecord

from ....utils import coalesce
from ....tool import Tool, ToolException
from ...common import ReadMethod, Base
from .datatype import Scaffold



# Maximal Phred Quality a base call can possibly achieve.
PHRED_MAX = 40



@dataclass
class RatedSequence :
    """
    Internal structure to store sequences and their quality score.
    """
    name : str
    seq : Optional[List[Base]]
    qual : Optional[List[float]]



class ShotgunSequencer (Tool) :
    """
    Function that, after setup, can take in a genome and perform whole-genome sequencing on it.
    The result will simulate a sequencing machine and provide a set of fragmented genome libraries
    that can later be assembled together to reconstruct the real genome of the host.

    Sequencing machines can be highly customizable and allow a multitude of parameters. Some
    considerations about the type of parameters and their default values are given below:

      - Usually a library size selection is done and this simulator will perform similar effects to
        those described under "Double-sided size selection and bead clean-up".
        A simplification is done and the blue graph in the link is converted to a normal distribution.
        https://emea.support.illumina.com/bulletins/2020/07/library-size-selection-using-sample-purification-beads.html

      - The read length of each fragment can also be configured, but a limit at roughly 150bp is set
        for each library, with some versions achieving up to 300bp. This limit is for each read.
        For example, with a value of 150bp, running the simulation on single-read method will yield
        libraries of 150bp each, whereas a paired-end method will yield two linked sequences
        of 150bp each.

      - Coverage of the extracted libraries can be controlled by reagents and number of cycles in
        the cloning step. The distribution of covered regions, however, depends on how the fragments
        attach to the flow cell prior to cluster forming. Attachment to the flow cell will follow
        random sampling from the available fragments. Given the amount of fragments this can be
        sampling to "sampling with reposition". The average coverage and read length; the genome
        size is known before-hand; and the number of reads will be calculated from all these values
        by using the Lander/Waterman equation.
        https://www.illumina.com/documents/products/technotes/technote_coverage_calculation.pdf

      - Quality of each call is difficult to simulate as it depends on multiple factors, such as:
        luminosity of dNTPs, signal-to-noise ratio, hardware of the sequencer and chemistry used.
        It defined by Illumina through empirical observations. For the simulation, an
        exponenetial distribution is used such, under default values, a Q30 score is achieved at
        97% of the times.
        https://www.illumina.com/Documents/products/technotes/technote_Q-Scores.pdf

    The sequences that are output will only contain the target region of each library.
    Adapter, index and tag regions are discarded.
    """

    # Average number of base pairs each resulting library will have.
    library_size_mean : float

    # Standard deviation in the length of resulting OpenReadFrames.
    library_size_sd : float

    # Method used when reading fragments.
    read_method : ReadMethod

    # Read length of each fragment.
    read_length : int

    # Average coverage of each base. Influences amount of frames that are returned.
    average_coverage : float

    # Beta scale parameter for the Exponential Distribution that determines the error rate.
    # It is the inverse of the rate parameter Lambda. (Beta = 1 / Lambda)
    # Higher values mean higher error rates.
    call_error_beta : float



    def __init__ (
        self,
        library_size_mean:float = 400,
        library_size_sd:float = 75,
        read_method:ReadMethod = 'single-read',
        read_length:int = 150,
        average_coverage:float = 10,
        call_error_beta:float = 2.85, # Estimation based on MiSeq graph in the TechNotes.
        name:Optional[str] = None,
        seed:Optional[int] = None
    ) :
        super().__init__( name=coalesce(name,"ShotgunSequencer"), seed=seed )
        self.library_size_mean = library_size_mean
        self.library_size_sd = library_size_sd
        self.read_method = read_method
        self.read_length = read_length
        self.average_coverage = average_coverage
        self.call_error_beta = call_error_beta



    def apply ( self, genome:Seq ) -> List[Scaffold] :
        rnd = self.make_generator()

        # Go for the approach of selecting randomly a start point and length, then check if it is
        # possible. Repeat if not.
        out_scaffolds:List[Scaffold] = []
        num_scaffolds = self.calc_num_scaffolds_obtained(genome)
        while len(out_scaffolds) < num_scaffolds :
            start = rnd.pick_integer( low=0, high=len(genome) ) # [low,high)
            library_size = int(rnd.pick_normal( self.library_size_mean, self.library_size_sd ))
            end = start + library_size

            # Repeat this sampling if it surpassed the possible range.
            if not ( end <= len(genome) ) :
                continue

            # Define the extracted the slices. R2 is only extracted if paired-end method is used.
            name_t = "BioLabSim.Sequencer id={} dir={}"
            r1 = RatedSequence( name=name_t.format(len(out_scaffolds),"R1"), seq=None, qual=None )
            r2 = RatedSequence( name=name_t.format(len(out_scaffolds),"R2"), seq=None, qual=None )

            # Extract the slices.
            read_size = min( library_size, self.read_length ) # Read is restricted by max read length.
            r1.seq = list(genome[ start : start + read_size ])
            if self.read_method == 'paired-end' :
                r2.seq = list(reversed(genome[ end - read_size : end ])) # R2 read in opposite direction

            # Apply quality estimation and substitution errors for both.
            for r in [ r1, r2 ] : # Same code for both possible reads.
                if r.seq != None : # Catch non-existent R2 reads.
                    r.qual = [ PHRED_MAX for _ in range(len(r.seq)) ] # Initialize quality.
                    for i in range(len(r.seq)) :

                        # Sample the error rate for each position from the exponential distribution.
                        q_error = rnd.pick_exponential( self.call_error_beta )
                        q_error = min( q_error, PHRED_MAX )
                        r.qual[i] = round(PHRED_MAX - q_error)

                        # Given the Phred Score, there is a probability a substitution error occurs.
                        if rnd.pick_uniform(0,1) < phred_to_prob(r.qual[i]) :
                            err_bases = [ b for b in ['A','C','G','T'] if b != r.seq[i] ]
                                # Always substitute for a base that is not the original.
                            sel_base = err_bases[ rnd.pick_integer(0,len(err_bases)) ]
                            r.seq[i] = sel_base

            # Create the scaffold and add it to the final results.
            cur_scaffold = Scaffold(
                expected_len=self.library_size_mean, # use `library_size` to store actual values
                r1_seqrecord=convert_to_seqrecord(r1),
                r2_seqrecord=convert_to_seqrecord(r2) if r2.seq is not None else None,
            )
            out_scaffolds.append(cur_scaffold)

        return out_scaffolds



    def calc_num_scaffolds_obtained ( self, genome:Seq ) -> int :
        """
        Number of frames that will be obtained. Calculated based on Lander/Waterman equation with
        regards to the current parameters and the host genome.
        This will account for the read method used. Paired-end methods will output half of it.
        """
        method_factor = 2 if self.read_method == 'paired-end' else 1
        return round(
            ( len(genome) * self.average_coverage )
            /
            ( self.read_length * method_factor )
        )



def phred_to_prob ( phred_score:float ) -> float :
    """ Convert a Phred score into a probability. """
    return 10 ** ( -phred_score / 10 )



def convert_to_seqrecord ( rated_seq:RatedSequence ) -> SeqRecord :
    """
    Convert an internal RatedSequence to the Seq that is commonly used throughout the library.
    """
    return SeqRecord(
        Seq( "".join(rated_seq.seq) ),
        id=rated_seq.name,
        name=rated_seq.name,
        description="",
        letter_annotations={ 'phred_quality': rated_seq.qual }
    )



import cobra
from cobra.io import read_sbml_model
import os
import wget
import numpy as np
import random
import matplotlib.pyplot as plt

def load_model():
    '''
    This function loads the E. coli core model
    args:
    None

    returns:
    model: cobra model
    '''
    # define boths paths where the model can be stored
    ModelFile = os.path.join('..', 'Data', 'e_coli_core.xml.gz')
    model = None

    # load the model from the first path where it is found
    if os.path.isfile(ModelFile):
            print(f'Loading existing file e_coli_core.xml.gz')
            # model = read_sbml_model(ModelFile)
            # return model
    else:
        print('Download of file e_coli_core.xml.gz from BIGG')
        # download the file from BIGG and save it in the `Data` directory
        wget.download('http://bigg.ucsd.edu/static/models/e_coli_core.xml.gz')
        # move the file to the `Data` directory
        os.rename('e_coli_core.xml.gz', ModelFile)
    model = read_sbml_model(ModelFile)
    return model

# substrates and products with carbon
def get_metabolites(model):
    '''
    This function gets the metabolites with carbon from the exchange reactions
    args:
    model: cobra model

    returns:
    metabolites: list of metabolites with carbon
    '''
    metabolites = []
    for rct in model.exchanges:
        for met in rct.metabolites:
            if 'C' in met.formula:
                metabolites.append(met.id)
            # # if met is co2 remove from list --> schönere Lösung finden!
            if met.id == 'co2_e':
                metabolites.remove(met.id)
    return metabolites

# check if reaction has positive product flux
def check_product(model, substrate, product):
    '''
    This function checks if the reaction has positive product flux
    args:
    model: cobra model
    substrate: str, substrate id
    product: str, product id

    returns:
    product: float, product flux
    '''
    with model:
        model.reactions.get_by_id('EX_glc__D_e').bounds = 0, 0
        model.reactions.get_by_id(f'EX_{substrate}').bounds = -10, 0
        # model objective to product
        model.objective = f'EX_{product}'
        product_max = model.slim_optimize()
        return product_max

# function to get sub and pro that fulfill the conditions
def get_sub_pro_pair(model, metabolites):
    '''
    This function gets a pair of metabolites from the exchange reactions with carbon, 
    where the corresponding reactions have positive growth rate and product flux.
    
    args:
    model: cobra model
    metabolites: list of metabolite IDs

    returns:
    tuple: pair of metabolite IDs (met_id1, met_id2) that fulfill the conditions
    '''
    # Select a random pair of metabolites from metabolites
    search_pair = True
    while search_pair:
        selected_pair = random.sample(metabolites, 2)
        # new pair if amount of carbon is the same
        diff_id = model.metabolites.get_by_id(selected_pair[0]).formula != model.metabolites.get_by_id(selected_pair[1]).formula
        # pair with more than 1 Carbon
        more_C = np.min([model.metabolites.get_by_id(selected_pair[0]).elements['C'], model.metabolites.get_by_id(selected_pair[1]).elements['C']]) > 1
        # check if product flux is positive
        product_bool = check_product(model, selected_pair[0], selected_pair[1]) > 0
        # search pair False, if all conditions are met
        if diff_id and more_C and product_bool:
            search_pair = False
        return selected_pair 
    else:
        print("Not enough metabolites found to form a pair.")  # Debugging output
        return None
    
# function to limit one random reaction in selected_pair by setting upper bound lower product flux
def limit_Rct(model, selected_pair, multiplier = 0.5):
    '''
    This function limits one radom reaction by setting the upper bound lower product flux
    args:
    model: cobra model
    selected_pair: list of met ids
    product: float, the product flux
    multiplier: float, the fraction of the product flux

    returns:
    model: cobra model with limited reaction
    '''
    # save reaction ID of all reactions from sub to pro in list
    my_Rct = f'EX_{selected_pair[0]}', f'EX_{selected_pair[1]}'  # Use f-string to get the reaction ID
    model.reactions.get_by_id(my_Rct[0]).lower_bound = -10
    model.reactions.get_by_id('EX_glc__D_e').bounds = 0, 0
    model.objective = my_Rct[1]
    solution = model.optimize()
    Rct = np.where(abs(round(solution.fluxes)) >= round(solution.objective_value))[0]
    EX_idx = [index for index, reaction in enumerate(model.reactions) if 'EX_' in reaction.id]
    valid_rct = list(set(Rct) - set(EX_idx))
    rct = random.choice(valid_rct)
    rct_id = model.reactions[rct].id
    if solution.fluxes[rct] > 0:
        model.reactions[rct].upper_bound = multiplier*solution.fluxes[rct]
    else:
        model.reactions[rct].lower_bound = multiplier*solution.fluxes[rct]
    return model, rct_id

# function to combine the functions above
def make_metabolite_combination(Student_ID):
    '''
    This function selects a pair of substrates and products from exchange metabolites and limits one reaction in the pathway. 
    args:
    Student_ID: int, the student ID
    
    returns:
    model: cobra model with limited reaction
    selected_pair: list of met ids
    product_max: float, maximum ofproduct flux
    product_lim: float, limited product flux
    rct: 
    '''
    random.seed(Student_ID)
    model = load_model()
    metabolites = get_metabolites(model)
    selected_pair = get_sub_pro_pair(model, metabolites)
    # variable with max product flux
    product = round(check_product(model, selected_pair[0], selected_pair[1]),2)
    # divide product_max by itself to get 1
    product_max = product / product
    model, rct = limit_Rct(model, selected_pair)
    # divide new product rate by product to get a value between 0 and 1
    product_lim = [round(model.slim_optimize(),2) / product]
    all_reactions = [f'{selected_pair[0]} --> {selected_pair[1]} limited']
    # product_min = [product_lim]
    return model, selected_pair, all_reactions, product, product_lim, rct

# function to set limited reaction back to original upper bounds in steps by factor 1.5
# beliebig oft durchführbar und wird immer um faktor 1.5 hochgesetzt
def optimize_reaction(model, rct_id, increase_factor = 1.5):
    '''
    This function sets the limited reaction back to the original upper bounds in steps by factor 1.5
    args:
    model: cobra model
    rct: str, reaction ID
    selected_pair: list of met ids

    returns: 
    model: cobra model with bounds set back
    '''
    solution = model.optimize()
    # increase upper bound 
    if solution.fluxes[model.reactions.get_by_id(rct_id).id] > 0:
        model.reactions.get_by_id(rct_id).upper_bound = model.reactions.get_by_id(rct_id).upper_bound * increase_factor
    # relax lower bound 
    else:
        model.reactions.get_by_id(rct_id).lower_bound = model.reactions.get_by_id(rct_id).lower_bound * increase_factor
    return model

# function to create bar chart of substrate-product pair and associated product flux    
def create_bar_chart(all_reactions, product_lim): # target_reaction):
    '''
    This function creates a bar chart of product flux
    args: 
    all_reactions: list of reaction IDs
    product_min: list of product fluxes
    
    returns:
    bar chart
    '''
    idx = range(len(all_reactions))
    plt.bar(idx, product_lim)
    plt.title('Product Flux for Substrate-Product Pair')
    plt.xlabel('Substrate-Product Combination')
    plt.ylabel('Product Flux [mmol/gDW/h]')
    plt.xticks(idx, all_reactions, rotation=90)
    plt.show()


"""
A Module is an extension of behaviour that can be attached to an Host.
"""

from __future__ import annotations
from abc import ABC
from typing import Optional, TYPE_CHECKING
from copy import copy

if TYPE_CHECKING:
    from ..base import Host

from .events import EventEmitter, EventLogger



class Module (ABC) :
    """
    A module usually contains 3 types of content:

      - host : The host this module belongs to.
      - deps* : Zero or more dependent modules this module is using.
      - params* : Zero or more params to build this module.

    Module creation happens can happen in 2 variations of steps:

      - make > bind > sync
      - copy > bind

    The steps are responsible for:

      - make : The initial params are set for a module created from scratch.
      - copy : The complete (copies have no sync step) params are set using another module as reference.
      - bind : The host and deps are set.
      - sync : The module executes a startup method that may include sending out events and
               using the dependent modules to initiate some data.
    """



    def __init__ ( self ) :
        """
        Will initialize the module by running either:
        - the make step if no reference object is given,
        - the copy step if a ref is given.

        This `__init__` should not be overriden. Instead, a derived class should override all of
        the `make`, `copy`, `bind` and `start` methods.
        """
        pass



    def make ( self ) -> None :
        """
        Make a new module from scratch with the help of arguments. This will only set the params.

        Extending Modules should have the following structure on their method:

        .. code-block:: python

            def make ( self, param_1, param_2, ... ) -> None :
                self.param_1 = param_1
                self.model_1 = load_model(param1)
                self.param_2 = param_2
        """
        raise ModuleException("Module does not implement `make`.")



    def copy ( self, ref:Module ) -> None :
        """
        Make a copy of the module params by using another similar Module as a reference.
        The code inside this method should provide a good copy where shallow and deep copies are
        used appropriately.

        Extending Modules should have the following structure on their method:

        .. code-block:: python

            def copy ( self, ref ) -> None :
                self.simple_param_1 = ref.simple_param_1
                self.complex_model_1 = copy(ref.complex_model_1)
        """
        raise ModuleException("Module does not implement `copy`.")



    def bind ( self, host:Host ) -> None :
        """
        Bind this module to its host and deps (dependent modules). Here, the event listeners should
        also be set.

        Extending Modules should have the following structure on their method:

        .. code-block:: python

            def bind ( self, host, req_mod_1, req_mod_2, ... ) -> None :
                self.host = host
                self.req_mod_1 = req_mod_1
                self.req_mod_2 = req_mod_2
                self.host.observe( EventTypeA, self.listen_event_a )
                self.host.observe( EventTypeB, self.listen_event_b )
        """
        raise ModuleException("Module does not implement `bind`.")



    def sync ( self, emit:EventEmitter, log:EventLogger ) -> None :
        """
        Run the sync procedure on this module. This usually means dispatching some events after the
        module is made and all other modules have been bound and are listening to event emitters.

        Extending Modules should have the following structure on their method:

        .. code-block:: python

            def sync ( self, emit, log ) -> None :
                log("ModuleX: start sync")
                emit( EventTypeA( event_args ) )
                for i in self.items :
                    emit( EventTypeB( event_args_i ) )

        """
        raise ModuleException("Module does not implement `sync`.")



class ModuleException (Exception) :
    """ Exception that is triggered when a module cannot be created. """
    pass



"""
Events that serve as communication for and between host modules.
"""

from abc import ABC
from typing import Callable


class Event (ABC) :
    """
    Events are changes to the Host that will be handled by the Modules.
    """
    pass


EventEmitter = Callable[[Event],None]  #:

EventLogger = Callable[[str],None]  #:




from ..sets.shotgun_sequencing.datatype import (
    get_consensus_from_overlap,
    estimate_from_overlap
)

from ..sets.shotgun_sequencing.evaluation import (
    calc_total_score,
    calc_sequence_score,
    evaluate_sequence
)

from ..sets.shotgun_sequencing.storage import (
    write_scaffolds_to_file
)

from ..sets.shotgun_sequencing.visualization import (
    print_scaffold_as_fastq,
    print_scaffold,
    print_assembly_evaluation,
    print_estimation_evaluation
)



"""
GenomeLibrary is a more complex version of GenomeList:
  - support for genome sequences
  - genes are located inside the sequence
"""


from __future__ import annotations
from typing import Callable, Optional, List, NamedTuple
from math import copysign
from copy import copy

# from numpy.random import Generator
from ...random import Generator
from cobra.core import Model as CobraModel
from cobra.core.solution import Solution as CobraSolution
from cobra.manipulation import delete_model_genes

from ...host import Host
from ...module import Module, ModuleException
from ...events import EventEmitter, EventLogger
from ..records.gene.stub_gene import StubGene
from ..all_events import InsertGeneEvent, RemoveGeneEvent, AlterGeneExpressionEvent
from ...extensions.utils.misc import Help_setDeactivateCExchanges, Help_getCarbonExchange, Help_CalcRate



class MetabolicFlux (Module) :
    """
    MetabolicFlux can handle cobrapy models and interfaces them with module events.

    This module can start with an non-existing model, which can then be integrated as an event.

    TODO: Its probably not elegant to allow non-existing models, but this makes it easier to add
    all genes in that model later on (as events). If we would be really pedantic and needed a
    strictly-existinging model we could implement an initialization step where a module calls
    multiple events on the host in order to "initialize" the module properly. To achieve that,
    modules themselves need to be able to generate events (right now only a Host can send events
    down to modules, events never go up the chain) and prevent infinite event loops. But maybe
    there are better alternatives altogether.
    """
    # Creator of random number generators. Host will bind this.
    make_generator: Callable[[],Generator]

    model: CobraModel

    # Since cobra models have limited methods to manage knocked-out genes, we need to keep track
    # of them in this list. A set of gene names.
    koed_genes: Set[str]


    def make ( self, model:CobraModel ) -> None :
        self.model = model
        self.model_tmp = self.model.copy()
        self.koed_genes = set()


    def copy ( self, ref:'MetabolicFlux' ) -> None :
        self.model = ref.model.copy()
        self.koed_genes = ref.koed_genes.copy()


    def bind ( self, host:Host ) -> None :
        self.make_generator = host.make_generator
        # host.observe( InsertGeneEvent, self.listen_insert_gene )
        # host.observe( RemoveGeneEvent, self.listen_remove_gene )
        # host.observe( AlterGeneExpressionEvent, self.listen_alter_gene_expression )


    def sync ( self, emit:EventEmitter, log:EventLogger ) -> None :
        log("MetabolicFlux: share model genes")
        for gene in self.model.genes : # Add all genes from the metabolic model.
            stub = StubGene( name=gene.name )
            emit( InsertGeneEvent(stub,locus=None) )

    def set_KinParameters ( self ) -> None :
        """
        Set kinetic parameters in the model. Each excahnge reaction is assigned with a random Vmax and Km.
        """

        CarbonExchanges = Help_getCarbonExchange(self.model)
        for ExRxn in CarbonExchanges:
            Vmax = round(self.make_generator().pick_uniform(low=2, high=20),1)
            Km = round(self.make_generator().pick_uniform(low=0.05, high=1),2)
            setattr(self.model.reactions.get_by_id(ExRxn), 'Vmax', Vmax)
            setattr(self.model.reactions.get_by_id(ExRxn), 'Km', Km)
            setattr(self.model_tmp.reactions.get_by_id(ExRxn), 'Vmax', Vmax)
            setattr(self.model_tmp.reactions.get_by_id(ExRxn), 'Km', Km)

    def reset( self ) -> None :
        self.model_tmp = self.model.copy()

    def optimize ( self ) -> CobraSolution :
        return self.model_tmp.optimize()
    
    def slim_optimize( self ) -> float :
        return round(self.model_tmp.slim_optimize(),2)
    
    def optimize_ReportExchanges ( self ) -> Tuple[ float, dict ] :
        solution = self.model_tmp.optimize()
        Exchanges = {}
        for rxn in self.model_tmp.exchanges:
            Exchanges[rxn.id] = round(solution.fluxes[rxn.id],2)
        return (round(solution.objective_value,2), Exchanges)

    def summary ( self ) -> str :
        return self.model_tmp.summary()

    def set_resetCarbonExchanges ( self , CarbonSubstrate:dict, Test:bool=False) -> None :
        """
        Set all carbon exchange reactions to zero uptake (lower_bound=0). Activate exchange reactions for input carbon sources.

        Inputs:
            CarbonSubstrate: dict, dictionary with carbon substrate exchange reactions as keys and their concentration in mmol/L as values
        Outputs:

        """
        if Test:
            print("MetabolicFlux: Test mode - not changing carbon exchange reactions")
            return
        # Deactivate all carbon exchange reactions 
        # using the original model to reset all changes
        self.model_tmp = Help_setDeactivateCExchanges(self.model_tmp) 
        # Activate exchange reactions for input carbon sources
        for ExRxn, Conc in CarbonSubstrate.items():
            Vmax = getattr(self.model_tmp.reactions.get_by_id(ExRxn), 'Vmax')
            Km = getattr(self.model_tmp.reactions.get_by_id(ExRxn), 'Km')
            UptakeRate = round(Help_CalcRate(Conc, Vmax, Km, Law='Michaelis-Menten'),1)
            setattr(self.model_tmp.reactions.get_by_id(ExRxn), 'lower_bound', -abs(UptakeRate))  # set uptake rate

  
    def listen_insert_gene ( self, event:InsertGeneEvent, emit:EventEmitter, log:EventLogger ) -> None :
        if event.gene.name in self.koed_genes :
            self.koed_genes.remove( event.gene.name )
            delete_model_genes( self.model, list(self.koed_genes), cumulative_deletions=False )
            log( "MetabolicFlux: reinserted gene={}".format(event.gene.name) )



    def listen_remove_gene ( self, event:RemoveGeneEvent, emit:EventEmitter, log:EventLogger ) -> None :
        """
        When a gene is removed we perform a knockout in the model.
        """
        found_genes = [ gene for gene in self.model.genes if gene.name == event.gene.name ]
        if len(found_genes) > 0 :
            self.koed_genes.add( event.gene.name )
            delete_model_genes( self.model, list(self.koed_genes), cumulative_deletions=False )
            log( "MetabolicFlux: knocked out gene={}".format(event.gene.name) )



    def listen_alter_gene_expression ( self, event:AlterGeneExpressionEvent, emit:EventEmitter, log:EventLogger ) -> None :
        """
        When a gene expression changes, we adapt the bounds.
        TODO: This method should be wrong. Please rewrite logic on how gene expression affects the
          metabolic model.
        """
        found_genes = [ gene for gene in self.model.genes if gene.name == event.gene.name ]
        if len(found_genes) > 0 :
            for gene in found_genes :
                for rct in gene.reactions :
                    rct.upper_bound = event.new_expression * copysign(1,rct.upper_bound)
                    log( "MetabolicFlux: changed upper_bound on reaction={} to {}".format(rct.id,rct.upper_bound) )



"""
Tools are long-standing objects that have to be initialized and that perform calculations on
other low-level values, such as sequences, matches, etc.

This class provides some common facilities such as stable randomization.
"""

from typing import List, Callable, NamedTuple, Optional, Type

from .events import Event
from .random import Generator, pick_seed
from .utils import coalesce



class ToolException (Exception) :
    """ Exception that is triggered when a Tool has a problem. """
    pass



class Tool :
    """ The Tool base class includes some basic blocks to enable a stable randomization. """

    name: str

    # By specifying a seed, the same code will produce the same results.
    rnd_seed: Optional[int]

    # A seed with an incremental counter provides a stable randomization were experiments can be
    # run multiple times with different results.
    rnd_counter: int



    def __init__ ( self, name:str = None, seed:Optional[int] = None ) :
        """
        Init is responsible to initialize the stable randomization.

        Code from derived classes should be:

            super().__init__( name, seed )
        """
        self.name = coalesce( name, 'unnamed' )
        self.rnd_seed = coalesce( seed, pick_seed() ) # by default, use a random seed
        self.rnd_counter = 0



    def make_generator ( self ) -> Generator :
        """ Construct a random number generator with the same seed stored in the host. """
        self.rnd_counter += 1
        return Generator( self.rnd_seed + self.rnd_counter )




from typing import Optional, List
from pathlib import Path

from Bio import SeqIO
from Bio.Seq import Seq
from Bio.SeqRecord import SeqRecord

from .datatype import Scaffold



def write_scaffolds_to_file ( scaffolds:List[Scaffold], r1_path:str, r2_path:Optional[str] = None ) :
    """
    Write a list of scaffolds into the filesystem as a FASTQ file.
    This method will detect paired-end sequences and store them in two different files in their
    reversed form.
    """

    # Resolve to the absolute paths from what was given from the user.
    r1_abspath:Path = Path(r1_path).resolve()
    r2_abspath:Optional[Path] = None
    if r2_path is not None :
        r2_abspath = Path(r2_path).resolve()

    # Fill both lists of SeqRecords that will be extracted from the scaffolds.
    r1_seqrecs:List[SeqRecord] = []
    r2_seqrecs:List[SeqRecord] = []
    for scaf in scaffolds :
        if scaf.r1_seqrecord is not None :
            r1_seqrecs.append( scaf.r1_seqrecord )
        if scaf.r2_seqrecord is not None :
            r2_seqrecs.append( scaf.r2_seqrecord )

    # Before writing the file, check that the file arguments are correct.
    if len(r2_seqrecs) > 0 and r2_abspath is None :
        raise Exception(
            "Tried to write_scaffolds_to_file. There is at least one paired-end sequence"
            " but no file to store it has been provided."
        )

    # Write the files to disk.
    r1_abspath.parent.mkdir(parents=True, exist_ok=True) # Build non-existing parent dirs.
    SeqIO.write( r1_seqrecs, r1_abspath, "fastq" )
    print("R1 file stored in: " + str(r1_abspath))
    if len(r2_seqrecs) > 0 :
        r2_abspath.parent.mkdir(parents=True, exist_ok=True) # Build non-existing parent dirs.
        SeqIO.write( r2_seqrecs, r2_abspath, "fastq" )
        print("R2 file stored in: " + str(r2_abspath))





from typing import Tuple, List, Callable
from collections import namedtuple
from copy import copy

import numpy as np
import pandas as pd

from ...host import Host
from ...module import Module
from ...outcome import Outcome
from ...events import EventEmitter, EventLogger
from ...random import Generator
from ..records.gene.gene import Gene
from ..utils.misc import Help_GrowthConstant, Growth_Maxrate, add_noise2
from .genome_expression import GenomeExpression
from ...experiment import ExperimentSettings



class GrowthBehaviour ( Module ) :

    # Creator of random number generators. Host will bind this.
    make_generator: Callable[[],Generator]

    # Dependent module Genome Expression holding the genes to express.
    genexpr: GenomeExpression

    opt_growth_temp: int
    max_biomass: int

    # Monod Parameters
    # Monod Substrate affinity
    Ks: float
    # Monod Yield coefficient
    Yxs: float
    # Monod Maximum growth rate
    umax: float



    def make ( self, opt_growth_temp:int, max_biomass:int, umax:float ) -> None :
        self.opt_growth_temp = opt_growth_temp
        self.max_biomass = max_biomass
        self.umax = umax

    def make2 ( self, opt_growth_temp:int, max_biomass:int , Ks:float, Yxs:float, k1:float, umax:float, OD2X ) -> None :
        self.opt_growth_temp = opt_growth_temp
        self.max_biomass = max_biomass
        self.Ks = Ks
        self.Yxs = Yxs
        self.k1 = k1
        self.umax = umax
        self.OD2X = OD2X


    def copy ( self, ref:'GrowthBehaviour' ) -> None :
        self.opt_growth_temp = ref.opt_growth_temp
        self.max_biomass = ref.max_biomass



    def bind ( self, host:Host, genexpr:GenomeExpression ) -> None :
        self.make_generator = host.make_generator # Pass bound method Host.make_generator
        self.genexpr = genexpr

    def bind2 ( self, host:Host ) -> None :
        self.make_generator = host.make_generator # Pass bound method Host.make_generator


    def sync ( self, emit:EventEmitter, log:EventLogger ) -> None :
        pass # Nothing to sync.



    def Make_TempGrowthExp ( self, CultTemps:list[int], exp_suc_rate:float ) -> Tuple[ pd.DataFrame, List[Tuple] ]:
        """
        TODO: Ideally, growth should return a table with enough information to contain the
        biomass results and to infer the loading times. Loading times is a synthetic construct
        that should be contained in a catalog Host, while the Module should only contain efficient
        calculations. The Host is an API for the User and may add interaction such as
        Resource management and loading times.

        TODO: If loading time should really be calculated here, then think about returning an
        additional record-oriented dataframe with columns (exp, temp, loading_len).

        TODO: I'm not sure oh the nuances but propose the following changes to make this conforming:
        - Remove the Experiment Success Rate and only use it at the "catalog.recexpsim.Host" level.
        - Make each simulation  call use a single temperature. (though, some calculations use all temps)
        - Each simulation returns a Series with the mass-over-time values.
        - Rename this method to "sim_cultivation" with renamed arguments.

        Returns
        =======
        [ mass_over_time_for_temp, pauses ]
            mass_over_time_for_temp
                Dataframe holding the mass over time for each temperature in a column.
            pauses
                List of tuples containing loading time information.

        """
        rnd = self.make_generator()

        CultTemps = np.array(CultTemps)
        Exp_Duration = 48

        capacity = self.max_biomass
        # the time of the half maximum population (inflection point) is calculated according to here:
        # https://opentextbc.ca/calculusv2openstax/chapter/the-logistic-equation/
        d_mult = 2  # we multiply the inflection point with 'd_mult' to increase cultivation time
        P0 = 0.1

        # determine time vector with maximum length:
        OptTemp = self.opt_growth_temp
        # Selecting the temperature that is most distant from the optimal temperature
        Temp_tmax = CultTemps[np.argmax(np.absolute(CultTemps-OptTemp))]
        # using the worst temperature to calculate lowest growth rate
        r_tmax = self.umax * Help_GrowthConstant(OptTemp, Temp_tmax)
        # using the worst temperature growth rate to compute longest simulation time, maximum set to 72 h
        duration_tmax = d_mult * 1/r_tmax * np.log((capacity - P0)/P0) + 1
        t_max = np.arange(np.minimum(Exp_Duration, duration_tmax))

        # create an empty DataFrame with t_max as first column
        col = []
        col.append('time [h]')
        for i in range(len(CultTemps)):
            col.append('exp.{} biomass conc. at {} °C'.format(i, (CultTemps[i])))

        df = pd.DataFrame(np.empty(shape=(len(t_max), len(CultTemps)+1), dtype=float), columns=col)
        df[:len(t_max)] = np.nan
        new_df = pd.DataFrame({'time [h]': t_max})
        df.update(new_df)

        PauseEntry = namedtuple('PauseEntry', 'exp loading_len')
        pauses = []

        # computing of biomass data and updating of DataFrame
        for i in range(len(CultTemps)):

            if rnd.pick_uniform(0,1) > exp_suc_rate: # TODO: Too nested. # experiment failure depending on investment to equipment
                r = self.umax * Help_GrowthConstant(OptTemp, CultTemps[i])
                # the result can reach very small values, which poses downstream problems, hence the lowest value is set to 0.05
                if r > 0.05: # under process conditions it might be realistic, source : https://www.thieme-connect.de/products/ebooks/pdf/10.1055/b-0034-10021.pdf
                    duration = Exp_Duration # d_mult * 1/r * np.log((capacity - P0)/P0) + 1
                else:
                    duration = Exp_Duration
                t = np.arange(np.minimum(Exp_Duration, duration))

                # biomass data is calculated according to https://en.wikipedia.host/wiki/Logistic_function
                mu = capacity / (1 + (capacity-P0) / P0 * np.exp(-r * t))
                sigma = 0.1*mu

                exp_TempGrowthExp = [rnd.pick_normal(mu[k], sigma[k]) for k in range(len(mu))]

                loading_len = len(t)
                exp = ' of exp.{} at {} °C'.format(i, (CultTemps[i]))
                pauses.append(PauseEntry( exp, loading_len ))

            else:
                mu = P0
                sigma = 0.08*mu
                exp_TempGrowthExp = [rnd.pick_normal(mu, sigma) for i in range(Exp_Duration)] # if cells haven't grown, the measurement is only continued for 6h

                loading_len = 7
                exp = ' of exp.{} at {} °C'.format(i, (CultTemps[i]))
                pauses.append(PauseEntry( exp, loading_len ))


            new_df = pd.DataFrame({'exp.{} biomass conc. at {} °C'.format(i, (CultTemps[i])): exp_TempGrowthExp})
            df.update(new_df)

        return ( df, pauses )

    def Make_MonodExperiment( self, CultTemps:list[int], exp_suc_rate:float ) -> Tuple[ pd.DataFrame, List[Tuple] ]:
        """

        Returns
        =======
        [ mass_over_time_for_temp, pauses ]
            mass_over_time_for_temp
                Dataframe holding the mass over time for each temperature in a column.
            pauses
                List of tuples containing loading time information.

        """
        rnd = self.make_generator()

        CultTemps = np.array(CultTemps)
        Exp_Duration = 48

        capacity = self.max_biomass
        # the time of the half maximum population (inflection point) is calculated according to here:
        # https://opentextbc.ca/calculusv2openstax/chapter/the-logistic-equation/
        d_mult = 2  # we multiply the inflection point with 'd_mult' to increase cultivation time
        P0 = 0.1

        # determine time vector with maximum length:
        OptTemp = self.opt_growth_temp
        # Selecting the temperature that is most distant from the optimal temperature
        Temp_tmax = CultTemps[np.argmax(np.absolute(CultTemps-OptTemp))]
        # using the worst temperature to calculate lowest growth rate
        r_tmax = Help_GrowthConstant(OptTemp, Temp_tmax)
        # using the worst temperature growth rate to compute longest simulation time, maximum set to 72 h
        duration_tmax = d_mult * 1/r_tmax * np.log((capacity - P0)/P0) + 1
        t_max = np.arange(np.minimum(Exp_Duration, duration_tmax))

        # create an empty DataFrame with t_max as first column
        col = []
        col.append('time [h]')
        for i in range(len(CultTemps)):
            col.append('exp.{} biomass conc. at {} °C'.format(i, (CultTemps[i])))

        df = pd.DataFrame(np.empty(shape=(len(t_max), len(CultTemps)+1), dtype=float), columns=col)
        df[:len(t_max)] = np.nan
        new_df = pd.DataFrame({'time [h]': t_max})
        df.update(new_df)

        PauseEntry = namedtuple('PauseEntry', 'exp loading_len')
        pauses = []

        # computing of biomass data and updating of DataFrame
        for i in range(len(CultTemps)):

            if rnd.pick_uniform(0,1) > exp_suc_rate: # TODO: Too nested. # experiment failure depending on investment to equipment
                r = Help_GrowthConstant(OptTemp, CultTemps[i])
                # the result can reach very small values, which poses downstream problems, hence the lowest value is set to 0.05
                if r > 0.05: # under process conditions it might be realistic, source : https://www.thieme-connect.de/products/ebooks/pdf/10.1055/b-0034-10021.pdf
                    duration = Exp_Duration # d_mult * 1/r * np.log((capacity - P0)/P0) + 1
                else:
                    duration = Exp_Duration
                t = np.arange(np.minimum(Exp_Duration, duration))

                # biomass data is calculated according to https://en.wikipedia.host/wiki/Logistic_function
                mu = capacity / (1 + (capacity-P0) / P0 * np.exp(-r * t))
                sigma = 0.1*mu

                exp_TempGrowthExp = [rnd.pick_normal(mu[k], sigma[k]) for k in range(len(mu))]

                loading_len = len(t)
                exp = ' of exp.{} at {} °C'.format(i, (CultTemps[i]))
                pauses.append(PauseEntry( exp, loading_len ))

            else:
                mu = P0
                sigma = 0.08*mu
                exp_TempGrowthExp = [rnd.pick_normal(mu, sigma) for i in range(Exp_Duration)] # if cells haven't grown, the measurement is only continued for 6h

                loading_len = 7
                exp = ' of exp.{} at {} °C'.format(i, (CultTemps[i]))
                pauses.append(PauseEntry( exp, loading_len ))


            new_df = pd.DataFrame({'exp.{} biomass conc. at {} °C'.format(i, (CultTemps[i])): exp_TempGrowthExp})
            df.update(new_df)

        return ( df, pauses )

    def Make_ProductionExperiment ( self, gene:Gene, CultTemp, GrowthRate, Biomass, ref_prom:str, accuracy_Test=.9 ) -> Outcome :
        """
        Return an outcome with the expression rate.
        """
        growth_const = Help_GrowthConstant(self.opt_growth_temp, self.opt_growth_temp)

        # testing whether the determined maximum biomass and the determined maximum growth rate are close to the actual ones
        if not (
            1 - np.abs(Biomass-self.max_biomass) / self.max_biomass > accuracy_Test
            and 1 - np.abs(GrowthRate-growth_const) / growth_const > accuracy_Test
        ) :
            return Outcome[float]( 0, 'Maximum biomass and/or maximum growth rate are incorrect.' )

        # Growth rate was only checked, for the calculation the rate resulting from the temperature is used
        r = Help_GrowthConstant(self.opt_growth_temp, CultTemp)
        GrowthMax = Growth_Maxrate(r, Biomass)
        PromStrength = self.genexpr.calc_prom_str( gene=gene, ref_prom=ref_prom )
        AbsRate = round(GrowthMax * PromStrength,2)
        FinalRelRate = round(AbsRate/self.Calc_MaxExpress(),2)
        return Outcome[float]( FinalRelRate )



    def Calc_MaxExpress (self) -> float :
        '''Function to calculate the maximum possible expression rate.'''
        BiomassMax = self.max_biomass
        OptTemp = self.opt_growth_temp
        factor = self.genexpr.infl_prom_str
        species_str = self.genexpr.species_prom_str
        MaximumPromoterStrength = round(species_str * factor,2)
        r = Help_GrowthConstant(self.opt_growth_temp, OptTemp)
        GrowthMax = Growth_Maxrate(r, BiomassMax)
        return round(GrowthMax * MaximumPromoterStrength,2)


    def Stream_TempGrowthExp ( self, ExperimentSettings:ExperimentSettings, Success_rate:float, Test:bool = False ) -> np.array :
        """
        Return a list of biomass values over time for a given experiment setting.
        The growth rate is adjusted according to the temperature using a growth constant function.
        The biomass values are perturbed with a normal distribution to simulate measurement noise.

        Arguments:
            ExperimentSetting: ExperimentSettings, settings for the experiment
        Outputs:
            List of biomass values over time        
        """
        if not Test:
            ExpFail_rand= 1#self.make_generator().pick_uniform(0,1)
            if ExpFail_rand > Success_rate:
                # calculating the variation of the growth rate based on the temperature difference
                GrowthRate = ExperimentSettings.GrowthRate * Help_GrowthConstant(self.opt_growth_temp, ExperimentSettings.Temperature)

                # biomass data is calculated according to https://en.wikipedia.org/wiki/Logistic_function
                K = ExperimentSettings.Capacity  # in gCDW/L
                P0 = ExperimentSettings.InitBiomass # in gCDW/L
                # calculate biomass over time
                Biomass = K / (1 + (K-P0) / P0 * np.exp(-GrowthRate * ExperimentSettings.SampleVector))

            else:
                # no growth, values around the initial biomass are selected from a normal distribution
                Biomass = np.tile(ExperimentSettings.InitBiomass, ExperimentSettings.SampleVector.shape)

            Biomass_perturbed = Biomass
            # np.array([np.random.normal(Biomass[k], sigma[k]) for k in range(len(Biomass))])

            return ( add_noise2(Biomass_perturbed, Success_rate) )

        # Unit testing mode
        else:
            # in test mode, generate 10 random  samples without input parameters
            Test = np.random.normal(0.4, 0.1, 10)
            return ( Test )


    def Stream_CalcMetConc ( self, Biomass:np.array, ExperimentSettings:ExperimentSettings, Success_rate:float) -> List[float] :
        """
        Return a list of metabolite concentrations over time for a given biomass list.
        The metabolite concentration is calculated based on the biomass and the yield coefficient.

        Arguments:
            Biomass: list of biomass values over time
            ExperimentSetting: ExperimentSettings, settings for the experiment
        Outputs:
            List of metabolite concentrations over time        
        """
        MetConc = abs(ExperimentSettings.CarbonConc - (Biomass - ExperimentSettings.InitBiomass) / ExperimentSettings.Yield)
        return ( add_noise2(MetConc, Success_rate) )




from ..sets.shotgun_sequencing.assembly import (
    ContigAssembler,
    GreedyContigAssembler,
    RandomContigAssembler
)

from ..sets.shotgun_sequencing.sequencing import (
    ShotgunSequencer
)



"""
Gene crafted by the user itself.
"""

from __future__ import annotations

from Bio.Seq import Seq

from .gene import Gene



class CraftedGene (Gene) :

    _name: str

    _orf: Seq

    _prom: Seq

    def __init__ ( self, name:str, orf:Seq, prom:Seq ) :
        super().__init__()
        self._name = name
        self._orf = orf
        self._prom = prom

    # def clone ( self ) -> CraftedGene :
    #     return CraftedGene(
    #         name=self._name,
    #         orf=self._orf, # Can pass because Seq is immutable.
    #         prom=self._prom, # Can pass because Seq is immutable.
    #     )

    @property
    def name ( self ) -> str :
        return self._name

    @property
    def orf ( self ) -> Seq :
        return self._orf

    @property
    def prom ( self ) -> Seq :
        return self._prom


    def make_variant ( self, orf_loc:int, orf_sub:Seq ) -> CraftedGene :
        """ TODO: Maybe its better to clone-then-mutate. """
        new_orf = self._orf[:orf_loc-1] + orf_sub + self._orf[orf_loc:] # orf_loc starts at 1
        return CraftedGene(
            name=self._name,
            orf=new_orf,
            prom=self._prom, # Can pass because Seq is immutable.
        )



"""
Utilities used for interaction with the student.
"""



def Help_Progressbar(n, loading_time, add):
    '''function for display of a loading bar, n: width of loading bar'''
    import sys
    import time

    loading = '.' * n
    for i in range(n+1):
        # this loop replaces each dot with a hash!
        print('\r%s progress{}: %3d percent'.format(add) % (loading, i*100/n), end='')
        loading = loading[:i] + '#' + loading[i+1:]
        time.sleep(loading_time)
    sys.stdout.write("\n")



"""
Experiment is the top-most scope. It includes all hosts and registries and acts as a global
namespace.
"""

import numpy as np
from typing import Set, Optional, Union
from abc import ABC

from numpy.random import SeedSequence

from .host import Host
from .random import Generator



class ExperimentException (Exception) :
    """ Exception that is triggered by an experiment. """
    pass



class Experiment (ABC) :
    """ The basic experiment will hold the hosts and a random-generator. """

    hosts: Set[Host]

    # The experiment will hold an internal random generator to allow repeatability.
    rnd_gen: Generator



    def __init__ ( self, seed:Optional[Union[int,SeedSequence]] = None ) :
        self.hosts = set()
        self.rnd_gen = Generator(seed)



    def bind_host ( self, host:Host ) -> None :
        """Keep the host accessible via the experiment."""
        self.hosts.add(host)



    # def create_host ( self, host_class:Type[Host], **kwargs ) -> None :
    #     host = host_class( exp=self, **kwargs )
    #     self.hosts.append( host )

class ExperimentSettings (ABC):
    """Class to hold experiment settings and parameters."""

    # definition of Experiment class with experiment settings
    ExperimentID: str # unique identifier for the experiment
    ExperimentType: str # e.g., "Batch", "Fed-Batch", "Continuous"
    Analytics: list # list of analytics to be performed, e.g., ['OD600', 'HPLC']
    HostName: str # name of the host organism
    Temperature: list # values in Celsius
    CultivationTime: int # default value in hours
    SamplingInterval: float # default sampling interval in hours
    InitBiomass: float # default initial biomass concentration in g/L
    MediumVolume: int # default value in mL
    MediumComposition: dict # dictionary to hold medium components and their concentrations
    CarbonID: str # identifier for the main carbon substrate, following BiGG nomenclature in the GSM model
    CarbonName: str # name of the main carbon substrate
    CarbonConc: float # concentration of the main carbon substrate in mM
    CarbonUptakeRate: float # uptake rate of the main carbon substrate in mmol/gCDW/h
    GrowthRate: float # specific growth rate in 1/h
    Yield: float # biomass yield in gCDW/mmol
    Capacity: int # maximum biomass concentration in g/L
    Results: str # file path to save experiment results

    def __init__(self) -> None :
        # initializing the experiment settings with empty values
        self.ExperimentID = ""
        self.ExperimentType = ""
        self.Analytics = []
        self.HostName = ""
        self.Temperature = []
        self.CultivationTime = 0
        self.SamplingInterval = 1.0
        self.InitBiomass = 0.0
        self.MediumVolume = 0
        self.MediumComposition = {}
        self.CarbonID = ""
        self.CarbonName = "" # name of the main carbon substrate
        self.CarbonConc = 0.0
        self.CarbonUptakeRate = 0.0
        self.GrowthRate = 0.0
        self.Yield = 0.0
        self.Capacity = 0
        self.Results = ""

    def set_SamplingVector(self) -> None :
        """Method to generate the sampling vector based on CultivationTime and SamplingInterval."""
        self.SampleVector = np.arange(0, self.CultivationTime + self.SamplingInterval, self.SamplingInterval)
        



import os
from pathlib import Path


# Global variable to see the directory where all data files are stored.
# TODO: In future, use an approach that works with setuptools and environment.
# DATADIR = Path(os.path.dirname(os.path.abspath(__file__))) / ".." / "data"
DATADIR = Path(os.path.dirname(os.path.abspath(__file__))) / "data" #:




from __future__ import annotations

from Bio.Seq import Seq

from .gene import Gene



class LocalizedGene (Gene) :
    """
    Gene that stores its sequence in a shared sequence.

    .. code-block::

        StartLoc       + PromLen       + OrfLen
            |----------------|--------------|

    TODO: Simple for now. Only has promoter and ORF which are adjacent to each other.
    """

    _name: str

    _seq: Seq

    _start_loc: int

    _prom_len: int

    _orf_len: int

    def __init__ ( self, name:str, seq:Seq, start_loc:int, prom_len:int, orf_len:int ) :
        super().__init__()
        self._name = name
        self._seq = seq
        self._start_loc = start_loc
        self._prom_len = prom_len
        self._orf_len = orf_len

    # def clone ( self ) -> LocalizedGene :
    #     return LocalizedGene(
    #         name=self.name,
    #         seq=self.seq, # Can pass because Seq is immutable.
    #         start_loc=self.start_loc,
    #         prom_len=self.prom_len,
    #         orf_len=self.orf_len,
    #     )

    @property
    def name ( self ) -> str :
        return self._name

    @property
    def orf ( self ) -> Seq :
        start = self._start_loc + self._prom_len
        end = start + self._orf_len
        return self._seq[start:end] # TODO: Test is end is include or exclusive

    @property
    def prom ( self ) -> Seq :
        start = self._start_loc
        end = start + self._prom_len
        return self._seq[start:end] # TODO: Test is end is include or exclusive



    @property
    def seq ( self ) -> Seq :
        return self._seq

    @property
    def start_loc ( self ) -> str :
        return self._start_loc

    @property
    def end_loc ( self ) -> str :
        return self._start_loc + self._prom_len + self._orf_len

    @property
    def prom_len ( self ) -> str :
        return self._prom_len

    @property
    def orf_len ( self ) -> str :
        return self._orf_len


    @seq.setter
    def seq ( self, value ) -> None:
        self._seq = value

    @start_loc.setter
    def start_loc ( self, value ) -> None:
        self._start_loc = value



"""
This file contains helper methods to visualize data structures and aid in debugging.
"""

from math import floor, ceil
from typing import Dict, List, Tuple

import numpy as np
from Bio.Seq import Seq

from .datatype import Scaffold, EstimatedSequence, LocalizedSequence, estimate_from_overlap
from .evaluation import calc_total_score, calc_sequence_score



def print_scaffold_as_fastq ( scaffold:Scaffold ) -> None :
    """
    Print the scaffold content as FASTQ.
    """
    print(scaffold.r1_seqrecord.format("fastq"))
    if scaffold.r2_seqrecord is not None :
        print(scaffold.r2_seqrecord.format("fastq"))



def print_scaffold ( scaffold:Scaffold ) -> None :
    """
    Visually print a scaffold as text.
    """
    r2_exists = scaffold.r2_seqrecord is not None

    # Calculate the gap size.
    gap = 0
    if scaffold.r2_seqrecord is None : # only R1 sequence [XX]---
        gap = scaffold.expected_len - len(scaffold.r1_seqrecord.seq)
    else : # both R1 and R2 sequence [XX]---[XX]
        gap = scaffold.expected_len - len(scaffold.r1_seqrecord.seq) - len(scaffold.r2_seqrecord)

    print( "| {r1s}~{lspad}(ca.{gap}){rspad}~{r2s}".format(
        r1s="".join(scaffold.r1_seqrecord.seq),
        r2s="".join(reversed(scaffold.r2_seqrecord)) if r2_exists else "",
        lspad="~" * ceil(gap / 2),
        rspad="~" * floor(gap / 2),
        gap=gap,
    ))



def numtochar ( num:int ) -> str :
    """
    Convert a number to a visual character.
    """
    if num == 0 :
        return "·"
    elif num >= 1 and num <= 9 :
        return str(round(num))
    elif num > 9 :
        return "#"
    return "?"



def print_evaluation_alignment (
    score:float, genome_start:int, genome:Seq,
    estseq:EstimatedSequence, locseqs:List[LocalizedSequence] = None
) -> None :
    """
    Print a visual representation of how alignment was evaluated.
    """

    # Calculate some parameters for nice formatting.
    bundle_start = 0
    bundle_end = len(estseq)
    window_start = min( bundle_start, genome_start )
    window_end = max( bundle_end, genome_start + len(genome) )

    # Print the heatmap.
    print( "-" * (13 + window_end - window_start) )
    for base in ['A','C','G','T'] :
        print("[heatmap:{}] {}{}{}".format(
            base,
            numtochar(0) * (bundle_start - window_start),
            "".join([ numtochar(floor(bp*10)) for bp in estseq.base_probs[base] ]),
            numtochar(0) * (window_end - bundle_end),
        ))

    # If localized sequences are given, print the coverage and each sequence.
    if locseqs is not None :

        locseq_start = np.amin([ locseq.locus for locseq in locseqs ])
        coverage = ""
        for pos in range( locseq_start, bundle_end + locseq_start ) :
            pos_cov = 0
            for locseq in locseqs :
                if locseq.locus <= pos < locseq.locus + len(locseq.sequence) :
                    pos_cov += 1
            coverage += numtochar(pos_cov)

        print( "-" * (13 + window_end - window_start) )
        print(" [coverage] {}{}{}".format(
            numtochar(0) * (bundle_start - window_start),
            coverage,
            numtochar(0) * (window_end - bundle_end),
        ))

        seq_scores = calc_sequence_score( locseqs, genome, genome_start )

        print( "-" * (13 + window_end - window_start) )
        for l_i, locseq in enumerate(locseqs) :
            print("  [seq:{:03}] {}{}{} ({:3.0f}%)".format(
                l_i,
                " " * (locseq.locus - locseq_start - window_start),
                str(locseq.sequence),
                " " * (window_end - (locseq.locus - locseq_start) - len(locseq.sequence)),
                seq_scores[l_i] * 100
            ))

    # Print the consensus sequence.
    consensus = estseq.as_consensus_sequence()
    print( "-" * (13 + window_end - window_start) )
    print("[consensus] {}{}".format(
        " " * (bundle_start - window_start),
        str(consensus)
    ))

    # Print the real genome.
    print( "-" * (13 + window_end - window_start) )
    print(" [real-gen] {}{}".format(
        " " * (genome_start - window_start),
        str(genome)
    ))

    # And at last, the final score.
    print( "-" * (13 + window_end - window_start) )
    print( "Final Score: {:.4f} ({:.1f}%)".format( score, score * 100 ) )



def print_assembly_evaluation ( loc_sequences:List[LocalizedSequence], genome:Seq ) -> None :
    """
    Given a list of localized sequences, print a visual representation of the evaluation.
    """
    est_sequence = estimate_from_overlap(loc_sequences)
    ( best_score, best_genome_start ) = calc_total_score( est_sequence, genome )
    print_evaluation_alignment(
        score=best_score, genome_start=best_genome_start, genome=genome,
        estseq=est_sequence, locseqs=loc_sequences
    )



def print_estimation_evaluation ( est_sequence:EstimatedSequence, genome:Seq ) -> None :
    """
    Given only an estimated sequence, print a visual representation of the evaluation.
    """
    ( best_score, best_genome_start ) = calc_total_score( est_sequence, genome )
    print_evaluation_alignment(
        score=best_score, genome_start=best_genome_start, genome=genome,
        estseq=est_sequence, locseqs=None
    )



"""
Methods that are not yet assigned to another place. They usually include methods before the
restructuring.
TODO: Almost all of the methods inside this file can me integrated into module functions.
"""
import os
import wget
import re
import numpy as np
import pandas as pd
import joblib
import pickle

from scipy.stats import norm
from Bio.Seq import Seq
from cobra.core.model import Model
from cobra.io import load_json_model

from ...random import pick_uniform
from ...outcome import Outcome
from .transform import list_onehot, list_integer
from ...extensions.common import BIGG_dict

def Sequence_ReferenceDistance(SeqObj, RefSeq):
    '''Returns the genetic sequence distance to a reference sequence.
    Input:
           SeqDF: list, the sequence in conventional letter format
    Output:
           SequenceDistance: float, genetic distances as determined from the sum of difference in bases divided by total base number, i.e. max difference is 1, identical sequence =0
    '''

    Num_Samp = len(SeqObj)
    SequenceDistance = np.sum([int(seq1 != seq2) for seq1,seq2 in zip(RefSeq, SeqObj)], dtype='float')/Num_Samp

    return SequenceDistance



def check_primer_integrity_and_recombination (Promoter:Seq, Primer:Seq, Tm:int, RefPromoter:Seq, OptPrimerLen:int) -> Outcome :
    '''
    Experiment to clone selected promoter. Return whether the experiment was successful.
    '''

    if Sequence_ReferenceDistance(Promoter, RefPromoter) > .4:
        return Outcome(False, 'Promoter sequence deviates too much from the given structure.')

    NaConc = 0.1 # 100 mM source: https://www.genelink.com/Literature/ps/R26-6400-MW.pdf (previous 50 mM: https://academic.oup.com/nar/article/18/21/6409/2388653)
    AllowDevi = 0.2 # allowed deviation
    Primer_Length = len(Primer)
    Primer_nC = Primer.count('C')
    Primer_nG = Primer.count('G')
    Primer_nA = Primer.count('A')
    Primer_nT = Primer.count('T')
    Primer_GC_content = ((Primer_nC + Primer_nG) / Primer_Length)*100 # unit needs to be percent

    if OptPrimerLen > 25:
        Primer_Tm = 81.5 + 16.6*np.log10(NaConc) + 0.41*Primer_GC_content - 600/Primer_Length # source: https://www.genelink.com/Literature/ps/R26-6400-MW.pdf (previous: https://core.ac.uk/download/pdf/35391868.pdf#page=190)
    else:
        Primer_Tm = (Primer_nT + Primer_nA)*2 + (Primer_nG + Primer_nC)*4
    # Product_Tm = 0.41*(Primer_GC_content) + 16.6*np.log10(NaConc) - 675/Product_Length
    # Ta_Opt = 0.3*Primer_Tm + 0.7*Product_Tm - 14.9
    # source Product_Tm und Ta: https://academic.oup.com/nar/article/18/21/6409/2388653
    # Product_Length would be the length of the promoter (40)? too small -> negative number comes out for Product_Tm
    print('Reference primer T:{}'.format(Primer_Tm))
    error = pick_uniform(-1,1)*0.1*Primer_Tm
#     error_2 = pick_uniform(-1,1)*0.1*Primer_Tm_2
    Primer_Tm_err = error + Primer_Tm
#     Primer_Tm_err_2 = error_2 + Primer_Tm_2

    DeviLen = np.absolute(OptPrimerLen - Primer_Length)/OptPrimerLen
    DeviTm = np.absolute(Primer_Tm_err - Tm)/Primer_Tm_err
#     DeviTm_2 = np.absolute(Primer_Tm_err_2 - Tm)/Primer_Tm_err_2
#     DeviTm = min(DeviTm_1, DeviTm_2)

    # create the complementary sequence of the primer to check for mistakes:
    PrimerComp = Primer.complement()

    if DeviLen <= AllowDevi and DeviTm <= AllowDevi/2 and Primer_Length <= 30 and PrimerComp == Promoter[:len(Primer)]:
        return Outcome(True)

    if not DeviLen <= AllowDevi :
        return Outcome(False, 'Primer length deviation too big.')
    if not DeviTm <= AllowDevi/2 :
        return Outcome(False, 'Temperature deviation too big.')
    if not Primer_Length <= 30 :
        return Outcome(False, 'Primer length too big.')
    if not PrimerComp == Promoter[:len(Primer)] :
        return Outcome(False, 'Primer not compatible with promoter.')

    return Outcome(False, 'Cloning failed')



def Help_PromoterStrength(
    PromSequence, RefPromoter:str, Scaler=1, Similarity_Thresh=.4, Regressor_File=None, AddParams_File=None
) -> float :
    '''
    TODO: Some arguments are missing.
    Expression of the recombinant protein.
        Arguments:
            Host:       class, contains optimal growth temperature, production phase
            Sequence:     string, Sequence for which to determine promoter strength
            Scaler:       int, multiplied to the regression result for higher values
            Predict_File: string, address of regression file
        Output:
            Expression: float, expression rate
    '''

    if Sequence_ReferenceDistance(PromSequence, RefPromoter) > Similarity_Thresh:
        Expression = 0
    else:
        Predictor = joblib.load(Regressor_File)
        Params = pickle.load(open(AddParams_File, 'rb'))
        Positions_removed = Params['Positions_removed']
        # Expr_Scaler = Params[Scaler_DictName]

        X = np.array(list_onehot(np.delete(list_integer(PromSequence),Positions_removed, axis=0))).reshape(1,-1)
        GC_cont = (PromSequence.count('G') + PromSequence.count('C'))/len(PromSequence)
        X = np.array([np.append(X,GC_cont)])
        Y = Predictor.predict(X)
        Expression = round(float(Y)*Scaler,3)

    return Expression



def Help_GrowthConstant(OptTemp, CultTemp, var=5):
    '''Function that generates the growth rate constant. The growth rate constant depends on the optimal growth temperature and the cultivation temperature. It is sampled from a Gaussian distribution with the mean at the optimal temperature and variance 1.
    Arguments:
        Opt_Temp: float, optimum growth temperature, mean of the Gaussian distribution
        Cult_Temp: float, cultivation temperature for which the growth constant is evaluated
        var: float, variance for the width of the Gaussian covering the optimal growth temperature
    Output:
        growth_rate_const: float, constant for use in logistic growth equation
    '''
    r_pdf = norm(OptTemp, var)
    # calculation of the growth rate constant, by picking the activity from a normal distribution

    growth_rate_const = r_pdf.pdf(CultTemp) / r_pdf.pdf(OptTemp)

    return growth_rate_const


def Calc_GSMMYield (model:Model, CarbExRxnID:str, MolarMass:float, UnitTest:bool=False) -> float :
    '''Function to calculate the growth rate for a genome scale metabolic model. The substrate and its concentration is used from the user input. 
    
    Inputs:
    '''

    if UnitTest:
        # for unit testing we want to have a fixed value
        Yield = 0.5
    else:
        solution = model.optimize()
        Yield = round(solution.objective_value / (abs(solution.fluxes[CarbExRxnID]) * MolarMass/1000),2)  # gDW / gSubstrate

    return Yield


def Growth_Maxrate(growth_rate_const, Biomass):
    '''
    TODO: Documentation is out of sync.
    The function calculates the maximum slope during growth.
        Arguments:
            Host: class, contains maximum biomass concentration as carrying capacity
            growth_rate_const: float, maximum growth rate constant
        Output:
            growth_rate_max: float, maximum growth rate
    '''
    # biomass checks
    # if Biomass > Mutant._Mutantmax_biomass or not Biomass:
    #     print('Error, no biomass was set or unexpected value or the maximum possible biomass was exceeded. Enter a value for the biomass again.')

    # Equation for calculating the maximum slope
    # https://www.tjmahr.com/anatomy-of-a-logistic-growth-curve/
    GrowthMax = Biomass * growth_rate_const / 4

    return GrowthMax

def Download_GSMM (Organism:str, ModelDir:str = 'Data') -> Model :
    '''Function to download a genome scale metabolic model from the BiGG database.
    
    Inputs:
        Organism: str, name of the organism in the BiGG database, e.g. 'e_coli_core'
        ModelFile: str, address to store the downloaded model
    Outputs:
        model: cobra Model, genome scale metabolic model
    '''

    ModelFile = os.path.join(ModelDir, f'{BIGG_dict[Organism]}.json')
    if os.path.isfile(ModelFile):
        # check if the file exists already
        model = load_json_model(ModelFile)
    else:
        # download the model from the BiGG database
        wget.download(f'http://bigg.ucsd.edu/static/models/{BIGG_dict[Organism]}.json')
        # move the file to the target directory
        os.rename(f'{BIGG_dict[Organism]}.json', ModelFile)
        model = load_json_model(ModelFile)

    return model

def Help_getCarbonExchange (model:Model) -> list :
    '''Function to get all carbon substrates from a genome scale metabolic model.
    
    Inputs:
        model: cobra Model, genome scale metabolic model
    Outputs:
        CarbonExchange: list, list of all carbon substrates in the model
    '''
    CarbonExchange = []
    for exchange in model.exchanges:
        metab = list(exchange.metabolites.keys())[0]
        # check if formula exist
        if bool(metab.formula):
            Carbon = bool(re.search(r'(?<![A-Za-z])C([A-Z]|\d)', metab.formula))
            if Carbon: # 'C' in metab.formula:
                CarbonExchange.append(exchange.id)

    return CarbonExchange

def Help_setDeactivateCExchanges (model:Model) -> Model :
    '''Function to deactivate all carbon exchange reactions.
    
    Inputs:
        model: cobra Model, genome scale metabolic model
        CarbonExchange: list, list of all carbon exchange reactions in the model
    Outputs:
        model: cobra Model, genome scale metabolic model with deactivated carbon exchange reactions
    '''
    CarbonExchange = Help_getCarbonExchange(model)

    for exchange in CarbonExchange:
        model.reactions.get_by_id(exchange).lower_bound = 0  # deactivate uptake

    return model

# def Help_setActivateCExchanges (model:Model, CarbonSubstrate:dict) -> Model :
#     '''Function to activate carbon exchange reactions with given uptake rates.
    
#     Inputs:
#         model: cobra Model, genome scale metabolic model
#         CarbonSubstrate: dict, dictionary with carbon substrates exchange reactions as keys and their uptake rates as values
#     Outputs:
#         model: cobra Model, genome scale metabolic model with activated carbon exchange reactions
#     '''

#     for ExRxn, rate in CarbonSubstrate.items():
#         model.reactions.get_by_id(ExRxn).lower_bound = -abs(rate)  # activate uptake

#     return model

def Help_CalcRate(S, Vmax, Km, Law:str='Michaelis-Menten') -> float :
    '''Function to calculate the carbon uptake rate based on Michaelis-Menten kinetics.
    
    Inputs:
        S: float, substrate concentration in mM
        Vmax: float, maximum uptake rate in mmol/gDW/h
        Km: float, Michaelis constant in mM
    Outputs:
        Rate: float, uptake rate in mmol/gDW/h
    '''
    if Law == 'Michaelis-Menten':
        Rate = (Vmax * S) / (Km + S)
    else:
        raise ValueError('Only Michaelis-Menten kinetics is implemented so far.')
    return Rate

def add_noise(data, noise):
    """
    Add noise to data.
    """
    # Each column get a different noise based on the average value of the column
    noise = noise * np.min(data, axis=0)   # np.mean(data, axis=0)  
    # ensure that there are no negative values
    return np.abs(data + np.random.normal(0, noise, data.shape))

def add_noise2(data, noise):
    '''
    Add noise to a vector, where each element is treated individually.
    '''
    return np.array([np.random.normal(data[k], noise*data[k]) for k in range(len(data))])



from __future__ import annotations
from abc import ABC


class Record ( ABC ) :
    """
    Record is a piece of information that can be indexed in a Registry.
    """
    pass

    # @abstractmethod
    # def clone ( self ) -> Record :
    #     pass




from typing import Optional
from dataclasses import dataclass

# from cobra.core import Model as CobraModel
from Bio.Seq import Seq

from ..events import Event
from .records.gene.gene import Gene


@dataclass
class InsertGeneEvent ( Event ) :
    gene: Gene
    locus: Optional[int]


@dataclass
class RemoveGeneEvent ( Event ) :
    gene: Gene


@dataclass
class AlterGenePromoterEvent ( Event ) :
    gene: Gene
    new_promoter: Seq
    # TODO: Not implemented in GeneList and GeneLibrary.


@dataclass
class AlterGeneExpressionEvent ( Event ) :
    gene: Gene
    new_expression: float




"""
The Gene component stores the:
 - ORF and Promoter sequences (enhancers in future?)
 - a name as designation
 - (annotations in future - all that what biopython brings to the table)
"""

from __future__ import annotations
from abc import ABC, abstractmethod

from Bio.Seq import Seq

from ....record import Record



class Gene ( Record, ABC ) :

    def __init__ ( self ) :
        pass

    # @abstractmethod
    # def clone ( self ) -> Gene :
    #     pass

    @property
    @abstractmethod
    def name ( self ) -> str :
        pass

    @property
    @abstractmethod
    def prom ( self ) -> Seq :
        pass

    @property
    @abstractmethod
    def orf ( self ) -> Seq :
        pass



"""
An host is a unit that can take in any number of host modules. It serves as the container
of all modules and allows for obverser-event communication between the modules. In addition, it is
capable of properly copying itself and it's modules. It also provides some facilities for
pseudo-random generation to allow deterministic generation.

In technical terms, the Host uses composition-over-inheritance to define all behaviours that
Modules may add to it. In addition to the usual forwarding methods, the Host is also an
observable that can be used by each module to communicate with other modules.
"""

from typing import List, Callable, NamedTuple, Optional, Type, TYPE_CHECKING
from abc import ABC

from .events import Event, EventEmitter, EventLogger
from .random import Generator, pick_seed
from .utils import coalesce



class HostException (Exception) :
    """ Exception that is triggered when a Host cannot be created. """
    pass



# The callback will execute provide an event to the listener an expect a message to add to
# the event log.
ListenerCallback = Callable[[Event,EventEmitter,EventLogger],Optional[str]]



class ListenerEntry ( NamedTuple ) :
    evtype: Type[Event]
    run: ListenerCallback



class Host (ABC) :
    """
    A host usually contains 3 types of content:

    - core params : Parameters that are used by all hosts.
    - modules : Zero or more modules inside the host.
    - extra params : Zero or more extra params used by the host extension.

    Host creation happens can happen in 2 variations of steps:

      - make > sync
      - copy

    The steps are responsible for:

      - make : The initial params are set for a host created from scratch. The extending host
      - copy : The complete (copies have no sync step) params are set using another host as reference.
      - sync : The host call out each module to sync.
    """

    name: str

    listeners: List[ListenerEntry]

    # The event_log holds small messages of all events that occured to the host.
    event_log: List[str]

    # By specifying a seed, the same code will produce the same results.
    rnd_seed: Optional[int]

    # A seed with an incremental counter provides a stable randomization were experiments can be
    # run multiple times with different results.
    rnd_counter: int

    # Number of clones made. This is used by the cloning method to generate new names.
    clone_counter: int



    def __init__ (
        self, ref:Optional['Host'] = None, name:str = None, seed:Optional[int] = None, **kwargs
    ) :
        """
        Init is responsible to initialize the base properties of an host, may it be new from
        scratch or by using another host as a reference.

        Extended Hosts should not override the __init__ method.
        """
        self.listeners = []
        self.event_log = []
        self.clone_counter = 0
        self.name = 'unnamed' # start unnamed and get a name later in this constructor
        self.rnd_seed = pick_seed() # by default, use a random seed
        self.rnd_counter = 0


        # Creating with ref puts some defaults on hierarchical names and stable seeds.
        if ref is not None :
            self.name, self.rnd_seed = ref.build_clone_attrs()

        # Constructor attributes always override all other values.
        if name is not None :
            self.name = name
        if seed is not None :
            self.rnd_seed = seed

        # Creation via "make > sync" is made if no ref is given. If a ref is given we perform "copy".
        # TODO: This is currently deactivated. Its a shorter style but for now prefer using the
        #   longer style because its very similar to how Modules are initialized.
        # if ref is None :
        #     self.make(**kwargs)
        #     self.sync()
        # else :
        #     self.copy(ref=ref)



    def make ( self ) -> None :
        """
        Make a new host from scratch with the help of arguments. Set possible params and modules.

        Extending Host should have the following structure on their method:

        .. code-block python::

            def make ( self, param_1, ... ) -> None :
                self.module_1 = ModuleTypeA()
                self.module_1.make( module_param_1=300, module_param_2=param_1 )
                self.module_1.bind( host=self )
        """
        raise HostException("Host does not implement `make`.")



    def copy ( self, ref:'Host' ) -> None :
        """
        Make a copy of the host params and modules by using another similar Host as a reference.
        The code inside this method should provide a good copy where shallow and deep copies are
        used appropriately.

        Extending Hosts should have the following structure on their method:

        .. code-block python::

            def copy ( self, ref ) -> None :
                self.module_1 = ModuleTypeA()
                self.module_1.copy( ref=self.genome )
                self.module_1.bind( host=self )
        """
        raise HostException("Host does not implement `copy`.")



    def sync ( self ) -> None :
        """
        Run the sync procedure on this host. It usually only calls the sync of each module using
        a helper method for that.

        Extending Hosts should have the following structure on their method:

        .. code-block python::

            def sync ( self ) -> None :
                self.sync_modules([ self.module_1, self.module_2, ... ])

        """
        raise HostException("Host does not implement `sync`.")



    def build_clone_attrs ( self ) -> [ str, int ] :
        """
        Generate attributes for a possible clone.

        Returns
        -------
        [ name, seed ]
            name: A hierarchical name is generated.
            seed: A stable seed is generated.
        """
        self.clone_counter += 1
        gen = self.make_generator()
        hierarchical_name = self.name + "." + str(self.clone_counter)
        stable_seed = gen.pick_seed() + self.clone_counter
        return [ hierarchical_name, stable_seed ]



    def log ( self, text:str, depth:Optional[int] = 1 ) -> None :
        """
        Log a message in the event log.
        """
        self.event_log.append( ("- " * depth) + text )



    def emit ( self, event: Event, chain: List[ListenerEntry] = [] ) -> None :
        """
        Trigger all listeners for a given Event.

        Uses a chain of listener entries to keep track of which has already been called
        and to abort loops that may result. Each listener callback may only execute once for each
        event chain.

        The Host also provides an EventEmitter to each listener callback in order to hide the functionality
        of listener chains from the modules. If they want to emit events they need to use the method
        provided by the host.
        """
        for le in self.listeners :
            if type(event) == le.evtype and le not in chain :
                # Run the listener with the emitted event and append to chain to keep track.
                new_chain = chain + [le]
                emit = lambda event : self.emit( event, new_chain )
                log = lambda text : self.log( text, len(new_chain)+1 )
                le.run( event, emit, log )



    def observe ( self, evtype: Type[Event], run: ListenerCallback ) -> None :
        """ Bind a new listener for a type of event. """
        self.listeners.append( ListenerEntry( evtype=evtype, run=run ) )



    def sync_modules ( self, modules:List['Module'] ) -> None :
        """
        Sync all modules as modules may emit some events or do load up calculations when they start.
        After creating all modules you should call this method with those modules.
        This cannot be included inside the module __init__ because all listeners have to be set.
        """
        emit = lambda event : self.emit( event, [] )
        log = lambda text : self.log( text )
        for mod in modules :
            mod.sync( emit, log )



    def make_generator ( self ) -> Generator :
        """ Construct a random number generator with the same seed stored in the host. """
        self.rnd_counter += 1
        return Generator( self.rnd_seed + self.rnd_counter )



    def print_event_log ( self ) -> None :
        print("Event Log for {}:".format(self.name))
        print("=" * (len(self.name) + 15))
        for ev in self.event_log :
            print(ev)



# Make these deep classes and methods available on the top-level.
# By convention, these are the public methods available for users of the library.

# We use a written `__init__.py` file at the top level which makes deep references into sub-modules.
# By doing this, all logic is inside this file and other __init__ files in the sub-folders can
# stay empty.

__version__='0.2.15'

from .config import DATADIR
from .experiment import Experiment, ExperimentException, ExperimentSettings
from .host import Host, HostException
from .tool import Tool, ToolException
from .outcome import SimulationException, Outcome, DataOutcome, DataWithPlotOutcome, combine_data
from .random import Generator
from .utils import alldef, coalesce, first

from .extensions.common import Base, PromoterSite, ReadMethod

from .extensions.all_events import (
    InsertGeneEvent, RemoveGeneEvent, AlterGenePromoterEvent, AlterGeneExpressionEvent
)

from .extensions.modules.genome_expression import GenomeExpression
from .extensions.modules.genome_library import GenomeLibrary
from .extensions.modules.genome_list import GenomeList
from .extensions.modules.growth_behaviour import GrowthBehaviour
from .extensions.modules.metabolic_flux import MetabolicFlux
from .extensions.modules.phenotype_size import PhenotypeSize

from .extensions.records.gene.gene import Gene
from .extensions.records.gene.crafted_gene import CraftedGene

from .extensions.tools.shotgun_sequencing import (
    ContigAssembler, GreedyContigAssembler, RandomContigAssembler, ShotgunSequencer
)

from .extensions.utils.shotgun_sequencing import (
    get_consensus_from_overlap, estimate_from_overlap, calc_total_score, calc_sequence_score,
    evaluate_sequence, write_scaffolds_to_file, print_scaffold_as_fastq, print_scaffold,
    print_assembly_evaluation, print_estimation_evaluation
)

from .catalog.RecExpSim import RecExperiment


"""
TODO: GenomeExpression needs to be reviewed.
"""

from typing import Union
from copy import copy

from Bio.Seq import Seq

from ...host import Host
from ...module import Module
from ...events import EventEmitter, EventLogger
from ..all_events import InsertGeneEvent, AlterGenePromoterEvent, AlterGeneExpressionEvent
from ..records.gene.gene import Gene
from ..utils.misc import Help_PromoterStrength
from .genome_library import GenomeLibrary
from .genome_list import GenomeList


CompatGenome = Union[GenomeLibrary,GenomeList]


class GenomeExpression ( Module ) :

    # Dependent module Genome Library holding the genes to express.
    genome: CompatGenome

    # Optimal primer length.
    opt_primer_len: int

    # Factor which influences the range of the promoter strength.
    # TODO: These characteristics maybe go elsewhere.
    infl_prom_str: float
    species_prom_str: float

    # Path to parameter files. TODO: use Paths.
    regressor_file: str
    addparams_file: str



    def make ( self,
        opt_primer_len:int,
        infl_prom_str:float,
        species_prom_str:float,
        regressor_file:str,
        addparams_file:str,
    ) -> None :
        self.opt_primer_len = opt_primer_len
        self.infl_prom_str = infl_prom_str
        self.species_prom_str = species_prom_str
        self.regressor_file = regressor_file
        self.addparams_file = addparams_file


    def copy ( self, ref:'GenomeExpression' ) -> None :
        self.opt_primer_len = ref.opt_primer_len
        self.infl_prom_str = ref.infl_prom_str
        self.species_prom_str = ref.species_prom_str
        self.regressor_file = ref.regressor_file
        self.addparams_file = ref.addparams_file



    def bind ( self, host:Host, genome:CompatGenome ) -> None :
        self.genome = genome
        host.observe( AlterGenePromoterEvent, self.listen_alter_gene_promoter )



    def sync ( self, emit:EventEmitter, log:EventLogger ) -> None :
        pass # Nothing to sync.



    def listen_alter_gene_promoter ( self, event:AlterGenePromoterEvent, emit:EventEmitter, log:EventLogger ) -> None :
        """ When a gene promoter is altered, we recalculate the expression and update it. """
        expr = self.calc_fast_wrong_prom_expr(event.new_promoter)
        log( "GenomeExpression: Changed gene={} to expression={}.".format(event.gene.name,expr) )
        emit(AlterGeneExpressionEvent( event.gene, expr ))



    def calc_fast_wrong_prom_expr ( self, prom:Seq ) -> float :
        """
        This method executes a wrong calculation of the promoter expression. It serves to
        demonstrate how editing a promoter can affect the expression value. That expression value
        can then change the metabolic models.
        TODO: Only for demonstration purpose of chained events.
        """
        # Use sum of deviations. Perfectly balanced counts will return an expression of 1.
        [a,c,g,t] = [ prom.count(base) for base in ["A","C","G","T"] ]
        avg = ( a + c + g + t ) / 4
        if avg == 0 : # On zero average, assume the sequence is undefined and always return 1.
            return 1.0
        dev = abs(a-avg) + abs(c-avg) + abs(g-avg) + abs(t-avg)
        expr = ( len(prom) * 2 - dev ) / ( len(prom) * 2 )
        return expr



    def calc_prom_str ( self, gene:Gene, ref_prom:str ) -> float :
        final_prom_str = float('NaN') # 0

        if gene in self.genome.genes :
            prom_str = Help_PromoterStrength(
                PromSequence=gene.prom,
                RefPromoter=ref_prom,
                Scaler=1,
                Similarity_Thresh=.4,
                Regressor_File=self.regressor_file,
                AddParams_File=self.addparams_file,
            )
            final_prom_str = round(prom_str * self.infl_prom_str, 2)

        return final_prom_str




from typing import Dict, List, Tuple
from math import inf

import numpy as np
from Bio.Seq import Seq

from ...common import Base
from .datatype import BaseProbDict, LocalizedSequence, EstimatedSequence



def calc_total_score ( estseq:EstimatedSequence, genome:Seq ) -> Tuple[float,int] :
    """
    Shift the genome around the heatmap and get the best overall match.
    Score is a simple addition of all ratios of the correct base, normalized as percentage.
    """

    gen_len = len(genome)
    est_len = len(estseq)
    best_score = -inf
    best_genome_start = 0
    for genome_start in range( -gen_len, est_len ) :
        score = 0
        for g_i in range(gen_len) : # g_i = index on genome sequence
            b_i = genome_start + g_i # b_i = map to an index in the bundle
            if 0 <= b_i < est_len : # only add score if it overlaps with the heatmap, 0 otherwise.
                gen_base = genome[g_i]
                hm_total = sum([ estseq.base_probs[base][b_i] for base in ['A','C','G','T'] ])
                if hm_total > 0 :
                    score += ( estseq.base_probs[gen_base][b_i] / hm_total )

        if score > best_score :
            best_score = score
            best_genome_start = genome_start

    return ( best_score / len(genome), best_genome_start )



def calc_sequence_score ( locseqs:List[LocalizedSequence], genome:Seq, genome_start:int ) -> List[float] :
    """
    Calculate the matching score of each singular localized sequence.
    """

    # genome_start is relative to min(locseqs.locus) when that is shifted to zero.
    locseq_start = int(np.amin([ locseq.locus for locseq in locseqs ]))
    rel_genome_start = genome_start + locseq_start

    seq_scores = [ 0 for _ in locseqs ]
    for l_i, locseq in enumerate(locseqs) :
        score = 0
        for s_i in range(len(locseq.sequence)) :
            g_i = locseq.locus + s_i - rel_genome_start
            if 0 <= g_i < len(genome) :
                if locseq.sequence[s_i] == genome[g_i] :
                    score += 1
        seq_scores[l_i] = score / len(locseq.sequence)

    return seq_scores



def evaluate_sequence ( estimated_sequence:EstimatedSequence, genome:Seq ) -> float :
    """
    Compares an estimated sequence with the real genome and returns the true positive rate.
    Return value is a number between 0 (worst score) and 1 (best score).
    """
    return calc_total_score( estimated_sequence, genome )[0]
